#include "CBondedParticle.h"

#include "Helpers.h"

using namespace std;
using namespace NApi;
using namespace NApiCm;
using namespace NApiCore;

const string CBondedParticle::PREFS_FILE            = "bonded_particle_prefs.txt";
const string CBondedParticle::BROKEN_BONDS_FILENAME = "broken_bonds.txt";

CBondedParticle::CBondedParticle() :
                 m_brokenBondsFileName(""),
                 m_brokenBondsFile(),
                 m_bondMakingHappening(false),
                 m_requestedBondTime(0.0),
                 m_bondingTime(0.0),
                 m_bondingTimestep(0.0)
{
    ;
}

CBondedParticle::~CBondedParticle()
{
    if (true == m_brokenBondsFile.is_open())
    {
        m_brokenBondsFile.close();
    }
}

void CBondedParticle::getPreferenceFileName(char prefFileName[FILE_PATH_MAX_LENGTH])
{
    // Copy pref file name
    strcpy(prefFileName, PREFS_FILE.c_str());
}

bool CBondedParticle::setup(NApiCore::IApiManager_1_0& apiManager,
                            const char                 prefFile[])
{
    ifstream prefsFile(prefFile);

    if(!prefsFile)
    {
        return false;
    }
    else
    {
        prefsFile >> m_requestedBondTime;

        double nNormalStiffness,
               nTangStiffness,
               nNormalStrength,
               nShearStrength,
               nBondDiskRadius;
        string str;

        while (prefsFile)
        {
            prefsFile >> str
                      >> nNormalStiffness
                      >> nTangStiffness
                      >> nNormalStrength
                      >> nShearStrength
                      >> nBondDiskRadius;

            string::size_type i (str.find (':'));
            string surfA = str.substr (0, i);
            str.erase (0, i + 1);
            string surfB = str;

            m_bondParameters.addBondParameters(surfA,
                                               surfB,
                                               CBondParameters(nNormalStiffness,
                                                               nTangStiffness,
                                                               nNormalStrength,
                                                               nShearStrength,
                                                               nBondDiskRadius));
        }

        // we want to write the broken bonds file in the same location
        // as the preference file, so extract the path from the pref file location
        string tmpPrefFile = prefFile;
        string::size_type idx = tmpPrefFile.find(PREFS_FILE);
        if(idx == string::npos)
        {
            return false;
        }

        m_brokenBondsFileName = tmpPrefFile.substr(0, idx);
        m_brokenBondsFileName += BROKEN_BONDS_FILENAME;

        // Open the broken bonds file

        m_brokenBondsFile.open(m_brokenBondsFileName.c_str(),
                               ios::out|ios::trunc );

        if(!m_brokenBondsFile)
        {
            return false;
        }
    }

    return true;
}

ECalculateResult CBondedParticle::calculateForce(
                                               double       time,
                                               double       timestep,
                                               int          elem1Id,
                                               const char   elem1Type[],
                                               double       elem1Mass,
                                               double       elem1ShearMod,
                                               double       elem1Poisson,
                                               double       elem1ContactCurvature,
                                               double       elem1PhysicalCurvature,
                                               double       elem1PosX,
                                               double       elem1PosY,
                                               double       elem1PosZ,
                                               double       elem1ComX,
                                               double       elem1ComY,
                                               double       elem1ComZ,
                                               double       elem1VelX,
                                               double       elem1VelY,
                                               double       elem1VelZ,
                                               double       elem1AngVelX,
                                               double       elem1AngVelY,
                                               double       elem1AngVelZ,
                                               double       elem1Charge,
                                               double       elem1WorkFunction,
                                               const double elem1Orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* elem1PropData,
                                               bool         elem2IsSurf,
                                               int          elem2Id,
                                               const char   elem2Type[],
                                               double       elem2Mass,
                                               double       elem2Area,
                                               double       elem2ShearMod,
                                               double       elem2Poisson,
                                               double       elem2ContactCurvature,
                                               double       elem2PhysicalCurvature,
                                               double       elem2PosX,
                                               double       elem2PosY,
                                               double       elem2PosZ,
                                               double       elem2ComX,
                                               double       elem2ComY,
                                               double       elem2ComZ,
                                               double       elem2VelX,
                                               double       elem2VelY,
                                               double       elem2VelZ,
                                               double       elem2AngVelX,
                                               double       elem2AngVelY,
                                               double       elem2AngVelZ,
                                               double       elem2Charge,
                                               double       elem2WorkFunction,
                                               const double elem2Orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* elem2PropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* contactPropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* simulationPropData,
                                               double       coeffRest,
                                               double       staticFriction,
                                               double       rollingFriction,
                                               double       contactPointX,
                                               double       contactPointY,
                                               double       contactPointZ,
                                               double       normalOverlap,
                                               double&      tangentialOverlapX,
                                               double&      tangentialOverlapY,
                                               double&      tangentialOverlapZ,
                                               double&      calculatedNormalForceX,
                                               double&      calculatedNormalForceY,
                                               double&      calculatedNormalForceZ,
                                               double&      calculatedUnsymNormalForceX,
                                               double&      calculatedUnsymNormalForceY,
                                               double&      calculatedUnsymNormalForceZ,
                                               double&      calculatedTangentialForceX,
                                               double&      calculatedTangentialForceY,
                                               double&      calculatedTangentialForceZ,
                                               double&      calculatedUnsymTangentialForceX,
                                               double&      calculatedUnsymTangentialForceY,
                                               double&      calculatedUnsymTangentialForceZ,
                                               double&      calculatedElem1AdditionalTorqueX,
                                               double&      calculatedElem1AdditionalTorqueY,
                                               double&      calculatedElem1AdditionalTorqueZ,
                                               double&      calculatedElem1UnsymAdditionalTorqueX,
                                               double&      calculatedElem1UnsymAdditionalTorqueY,
                                               double&      calculatedElem1UnsymAdditionalTorqueZ,
                                               double&      calculatedElem2AdditionalTorqueX,
                                               double&      calculatedElem2AdditionalTorqueY,
                                               double&      calculatedElem2AdditionalTorqueZ,
                                               double&      calculatedElem2UnsymAdditionalTorqueX,
                                               double&      calculatedElem2UnsymAdditionalTorqueY,
                                               double&      calculatedElem2UnsymAdditionalTorqueZ,
                                               double&      calculatedChargeMovedToElem1)
{
    // Some temporary variables
    CSimple3DVector F_bond_n,
                    F_bond_t,
                    F_HM_n,
                    F_HM_t,
                    F_damping_n,
                    F_damping_t,
                    F_total_n,
                    F_total_t,
                    T_bond_1,
                    T_bond_2,
                    T_damping_1,
                    T_damping_2,
                    T_total_1,
                    T_total_2;

    /************************************************************************/
    /* Making / obtaining bonds from records                                */
    /************************************************************************/
    CBond elasticBond;

    bool bBondExists = false;

    if (true == elem2IsSurf)
    {
        //initial assignment
        bool bFormBond = false;
        if( !m_bondMakingHappening && time > m_requestedBondTime)
        {
            m_bondMakingHappening = true;
            m_bondingTime = time;
            m_bondingTimestep = timestep;
            bFormBond = true;
        }
        else if(m_bondMakingHappening && time < ( m_bondingTime + m_bondingTimestep / 2.0) )
        {
            bFormBond = true;
        }

        if(bFormBond)
        {
            CBond thisBond(elem1Id, elem2Id);
            m_bonds.addObject(elem1Id,elem2Id, thisBond);
            tangentialOverlapX = tangentialOverlapY = tangentialOverlapZ = 0.0;
            bBondExists = true;
        }
        else
        {
            if(m_bonds.contains(elem1Id, elem2Id))
            {
                bBondExists = true;
                elasticBond = m_bonds.getObject(elem1Id, elem2Id);
            }
        }
    }

    /************************************************************************/
    /* Predefining useful information on contact                            */
    /************************************************************************/

    /* The unit vector from element 1 to the contact point */
    CSimple3DPoint contactPoint     = CSimple3DPoint(contactPointX, contactPointY, contactPointZ);
    CSimple3DVector unitCPVect      = contactPoint - CSimple3DPoint(elem1PosX, elem1PosY, elem1PosZ) ;
    unitCPVect.normalise();

    /* Calculate the relative velocities of the elements*/
    CSimple3DVector vel1(elem1VelX, elem1VelY, elem1VelZ);
    CSimple3DVector vel2(elem2VelX, elem2VelY, elem2VelZ);
    CSimple3DVector relVel        =  vel1 - vel2;
    CSimple3DVector relVel_n      = unitCPVect * unitCPVect.dot( relVel);
    CSimple3DVector relVel_t      = relVel - relVel_n;
    CSimple3DVector angVel1(elem1AngVelX, elem1AngVelY, elem1AngVelZ);
    CSimple3DVector angVel2(elem2AngVelX, elem2AngVelY, elem2AngVelZ);

    /*Equivalent radii & mass */
    double nEquivRadius     =  elem1PhysicalCurvature * elem2PhysicalCurvature / (elem1PhysicalCurvature + elem2PhysicalCurvature );
    double nEquivMass       = elem1Mass * elem2Mass / (elem1Mass + elem2Mass);

    /*Effective Young's Modulus*/
    double nYoungsMod1      = 2 * (1.0 + elem1Poisson) * elem1ShearMod;
    double nYoungsMod2      = 2 * (1.0 + elem2Poisson) * elem2ShearMod;

    /*Equivalent Young's & shear modulus*/
    double nEquivYoungsMod  =  ( 1 - pow(elem1Poisson, 2) ) / nYoungsMod1
                            + ( 1 - pow(elem2Poisson, 2) ) / nYoungsMod2;
    nEquivYoungsMod         = 1.0 / nEquivYoungsMod;

    double nEquivShearMod   =   ( 2.0 - elem1Poisson ) /  elem1ShearMod
                            + ( 2.0 - elem2Poisson ) /  elem2ShearMod;

    nEquivShearMod          = 1.0 / nEquivShearMod;

    /* Tangential Overlap */
    CSimple3DVector nOverlap_t(tangentialOverlapX, tangentialOverlapY, tangentialOverlapZ);

    /************************************************************************/
    /* Bonded Particle Calculation                                          */
    /************************************************************************/
    if(bBondExists && !elasticBond.getBroken())
    {
        CBondParameters bondParams = m_bondParameters.getBondParameters(elem1Type, elem2Type);

        double nArea = PI * bondParams.m_nBondDiskRadius * bondParams.m_nBondDiskRadius;
        CSimple3DVector ndOverlap_n = relVel_n * timestep;
        CSimple3DVector ndOverlap_t = relVel_t * timestep;

        CSimple3DVector dF_n = -ndOverlap_n * bondParams.m_nNormalStiffness * nArea;
        CSimple3DVector dF_t = -ndOverlap_t * bondParams.m_nTangStiffness * nArea;

        double J = 0.5 * PI * pow(bondParams.m_nBondDiskRadius,4);
        double I = J / 2.0;

        CSimple3DVector angle  = (angVel1 -angVel2) * timestep;
        CSimple3DVector normalAngle = unitCPVect * unitCPVect.dot(angle);
        CSimple3DVector tangAngle = angle - normalAngle;

        CSimple3DVector dT_n = - normalAngle * J * bondParams.m_nTangStiffness;
        CSimple3DVector dT_t = - tangAngle * I * bondParams.m_nNormalStiffness;

        F_bond_n = elasticBond.getForceNormal() + dF_n;
        F_bond_t = elasticBond.getForceTangential() + dF_t;

        CSimple3DVector T_n = elasticBond.getTorqueNormal() + dT_n;
        CSimple3DVector T_t = elasticBond.getTorqueTangential() + dT_t;

        /* check for fracture */
        double maxStress = -F_bond_n.length() / nArea + T_t.length() / I * bondParams.m_nBondDiskRadius;
        double maxTorque = F_bond_t.length() / nArea + T_n.length() / J * bondParams.m_nBondDiskRadius;
        m_bonds.removeObject(elem1Id, elem2Id);

        if(fabs(maxStress) <= bondParams.m_nNormalStrength && fabs(maxTorque) <= bondParams.m_nShearStrength)
        {
            T_bond_1 = T_n + T_t;
            T_bond_2 = -T_bond_1;

            elasticBond.getForceNormal() = F_bond_n;
            elasticBond.getForceTangential() = F_bond_t;
            elasticBond.getTorqueNormal() = T_n;
            elasticBond.getTorqueTangential() = T_t;
        }
        else
        {
            F_bond_n = CSimple3DVector();
            F_bond_t = CSimple3DVector();
            elasticBond.setBroken(true);

            m_brokenBondsFile << "id1: " << elem1Id
                              << " pos: x=" << elem1PosX << ", y=" << elem1PosY << ", z=" << elem1PosZ
                              << " id2: " << elem2Id
                              << " pos: x=" << elem2PosX << ", y=" << elem2PosY << ", z=" << elem2PosZ
                              << " time: " << time << endl;
        }

        m_bonds.addObject(elem1Id, elem2Id, elasticBond);
    }

    /************************************************************************/
    /* Hertz-Mindlin Calculation                                            */
    /************************************************************************/

    //Although the bond might exist, we may not have actual physical overlap.
    double nPhysicalOverlap = normalOverlap;
    nPhysicalOverlap += elem1PhysicalCurvature - elem1ContactCurvature;

    if(true == elem2IsSurf)
    {
        nPhysicalOverlap += elem2PhysicalCurvature - elem2ContactCurvature;
    }

    if( nPhysicalOverlap > 0)
    {
        /*********************************************/
        /* Correcting for physical vs contact radius */
        /*********************************************/

        //Tangential overlap.may also be wrong
        nOverlap_t -= relVel_t * timestep;

        //the normal velocity is unaffected, but the tangential one might be different
        //this approximation assumes the particles are spheres.
        vel1 += angVel1.cross(unitCPVect*(elem1PhysicalCurvature - elem1ContactCurvature));

        if(true == elem2IsSurf)
        {
            vel2 += angVel2.cross(-unitCPVect * (elem2PhysicalCurvature - elem2ContactCurvature));
        }

        relVel =  vel1 - vel2;
        relVel_n  = unitCPVect * unitCPVect.dot( relVel);
        relVel_t = relVel - relVel_n;

        nOverlap_t += relVel_t * timestep;

        /*****************************/
        /* Normal Calculation        */
        /*****************************/

        double K_n   = 4.0 / 3.0 * nEquivYoungsMod * sqrt(nEquivRadius);
        F_HM_n  = - unitCPVect * K_n * pow(nPhysicalOverlap, 1.5);

        /*****************************/
        /* Tangential Calculation    */
        /*****************************/

        double S_t =  8.0 * nEquivShearMod * sqrt(nEquivRadius * nPhysicalOverlap);
        F_HM_t = -nOverlap_t * S_t;

        /*****************************/
        /* Damping/Friction Calculation*/
        /*****************************/

        double B = 0.0;
        if(coeffRest > 0.0)
        {
            double myLog = log(coeffRest);
            B = - myLog/ sqrt(myLog * myLog + PI * PI);
        }

        double S_n = 2.0 * nEquivYoungsMod * sqrt(nEquivRadius * nPhysicalOverlap);
        F_damping_n = unitCPVect * 2 * sqrt(5.0 / 6.0) * B * sqrt(S_n * nEquivMass) *  relVel_n.length();

        /*Are we in a loading situation?*/
        if(relVel_n.dot( unitCPVect) > 0.0)
        {
            F_damping_n = -F_damping_n;
        }

        if(F_HM_t.length() > F_HM_n.length() * staticFriction)
        {
            F_HM_t = F_HM_t * F_HM_n.length() * staticFriction / F_HM_t.length();
            /* slippage has occurred so the tangential overlap is reduced a bit */
            nOverlap_t = -F_HM_t / S_t;

            /* at this point we get energy loss from the sliding */
            F_damping_t = F_HM_t;
            /* little trick so the formulae works below */
            F_HM_t = CSimple3DVector();
        }
        else
        {
            /* at this point we get energy loss from the damping */
            F_damping_t = - relVel_t * 2 * sqrt(5.0 / 6.0) * B * sqrt(S_t * nEquivMass);
        }

        /*****************************/
        /* Rolling Friction          */
        /*****************************/

        /*Only relevant if actually rolling*/
        if(!isZero(angVel1.lengthSquared()))
        {
            CSimple3DVector torque1 = angVel1;
            torque1.normalise();
            torque1 *= -F_HM_n.length() * elem1PhysicalCurvature * rollingFriction;
            T_damping_1 = torque1;
        }

        if(!isZero(angVel2.lengthSquared()))
        {
            CSimple3DVector torque2 = angVel2;
            torque2.normalise();
            torque2 *= -F_HM_n.length() * elem2PhysicalCurvature * rollingFriction;
            T_damping_2 = torque2;
        }
    }

    /****************************************/
    /* Fill in parameters we were passed in */
    /****************************************/
    F_total_n = F_bond_n + F_HM_n + F_damping_n;
    F_total_t = F_bond_t + F_HM_t + F_damping_t;
    T_total_1 = T_damping_1;
    T_total_2 = T_damping_2;

    calculatedNormalForceX = F_total_n.dx();
    calculatedNormalForceY = F_total_n.dy();
    calculatedNormalForceZ = F_total_n.dz();

    calculatedUnsymNormalForceX = F_damping_n.dx();
    calculatedUnsymNormalForceY = F_damping_n.dy();
    calculatedUnsymNormalForceZ = F_damping_n.dz();

    calculatedTangentialForceX = F_total_t.dx();
    calculatedTangentialForceY = F_total_t.dy();
    calculatedTangentialForceZ = F_total_t.dz();

    calculatedUnsymTangentialForceX = F_damping_t.dx();
    calculatedUnsymTangentialForceY = F_damping_t.dy();
    calculatedUnsymTangentialForceZ = F_damping_t.dz();

    tangentialOverlapX = nOverlap_t.dx();
    tangentialOverlapY = nOverlap_t.dy();
    tangentialOverlapZ = nOverlap_t.dz();

    calculatedElem1AdditionalTorqueX = T_total_1.dx();
    calculatedElem1AdditionalTorqueY = T_total_1.dy();
    calculatedElem1AdditionalTorqueZ = T_total_1.dz();

    calculatedElem1UnsymAdditionalTorqueX = T_damping_1.dx();
    calculatedElem1UnsymAdditionalTorqueY = T_damping_1.dy();
    calculatedElem1UnsymAdditionalTorqueZ = T_damping_1.dz();

    calculatedElem2AdditionalTorqueX = T_total_2.dx();
    calculatedElem2AdditionalTorqueY = T_total_2.dy();
    calculatedElem2AdditionalTorqueZ = T_total_2.dz();

    calculatedElem2UnsymAdditionalTorqueX = T_damping_2.dx();
    calculatedElem2UnsymAdditionalTorqueY = T_damping_2.dy();
    calculatedElem2UnsymAdditionalTorqueZ = T_damping_2.dz();

    return eSuccess;
}