#if !defined(CBondedParticle_h)
#define CBondedParticle_h

#include <string>
#include <fstream>

#include "IPluginContactModelV2_1_0.h"
#include "CBondList.h"
#include "CBondParametersList.h"

/**
 * This class provides an implementation of IPluginContactModelV2_1_0
 * That adds performs hertz mindlin contact force calculation
 * with added bonded particle physics
 *
 * This plugin reads a config file from the same directory as
 * it is stored in.  The file format is a single line containing
 *     The time at which bonds should be formed
 *
 * Followed by any number of lines defining bond types:
 *     type1:type_2 normal_stiffness
 *                  tangential_stiffness
 *                  normal_strength
 *                  shear_strength
 *                  bond_disk_radius;
 *
 * NOTE: This version of the hertz-midnlin & bonded particle code is
 * provided only as an example of how to create a contact model
 * plugin and is not updated on a regular basis.
 */
class CBondedParticle : public NApiCm::IPluginContactModelV2_1_0
{
    public:
        /**
         * Name of the preferences file to load bond information
         * from
         */
        static const std::string PREFS_FILE;

        /**
         * Name of the temp file to hold broken bond information
         */
        static const std::string BROKEN_BONDS_FILENAME;

        /**
         * Constructor, does nothing
         */
        CBondedParticle();

        /**
         * Destructor, does nothing
         */
        virtual ~CBondedParticle();

        /**
         * Sets prefFileName to PREFS_FILE
         *
         *
         * Implementation of method from IPluginContactModelV2_1_0
         */
        virtual void getPreferenceFileName(char prefFileName[NApi::FILE_PATH_MAX_LENGTH]);

        /**
         * Reads config file
         *
         * Implementation of method from IPluginContactModelV2_1_0
         */
        virtual bool setup(NApiCore::IApiManager_1_0& apiManager,
                           const char                 prefFile[]);

        /**
         * Performs Hertz Mindlin contact calculations with added
         * bonded particle mechanics
         *
         * Implementation of method from IPluginContactModelV2_1_0
         */
        virtual NApi::ECalculateResult calculateForce(
                                               double       time,
                                               double       timestep,
                                               int          elem1Id,
                                               const char   elem1Type[],
                                               double       elem1Mass,
                                               double       elem1ShearMod,
                                               double       elem1Poisson,
                                               double       elem1ContactCurvature,
                                               double       elem1PhysicalCurvature,
                                               double       elem1PosX,
                                               double       elem1PosY,
                                               double       elem1PosZ,
                                               double       elem1ComX,
                                               double       elem1ComY,
                                               double       elem1ComZ,
                                               double       elem1VelX,
                                               double       elem1VelY,
                                               double       elem1VelZ,
                                               double       elem1AngVelX,
                                               double       elem1AngVelY,
                                               double       elem1AngVelZ,
                                               double       elem1Charge,
                                               double       elem1WorkFunction,
                                               const double elem1Orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* elem1PropData,
                                               bool         elem2IsSurf,
                                               int          elem2Id,
                                               const char   elem2Type[],
                                               double       elem2Mass,
                                               double       elem2Area,
                                               double       elem2ShearMod,
                                               double       elem2Poisson,
                                               double       elem2ContactCurvature,
                                               double       elem2PhysicalCurvature,
                                               double       elem2PosX,
                                               double       elem2PosY,
                                               double       elem2PosZ,
                                               double       elem2ComX,
                                               double       elem2ComY,
                                               double       elem2ComZ,
                                               double       elem2VelX,
                                               double       elem2VelY,
                                               double       elem2VelZ,
                                               double       elem2AngVelX,
                                               double       elem2AngVelY,
                                               double       elem2AngVelZ,
                                               double       elem2Charge,
                                               double       elem2WorkFunction,
                                               const double elem2Orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* elem2PropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* contactPropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* simulationPropData,
                                               double       coeffRest,
                                               double       staticFriction,
                                               double       rollingFriction,
                                               double       contactPointX,
                                               double       contactPointY,
                                               double       contactPointZ,
                                               double       normalOverlap,
                                               double&      tangentialOverlapX,
                                               double&      tangentialOverlapY,
                                               double&      tangentialOverlapZ,
                                               double&      calculatedNormalForceX,
                                               double&      calculatedNormalForceY,
                                               double&      calculatedNormalForceZ,
                                               double&      calculatedUnsymNormalForceX,
                                               double&      calculatedUnsymNormalForceY,
                                               double&      calculatedUnsymNormalForceZ,
                                               double&      calculatedTangentialForceX,
                                               double&      calculatedTangentialForceY,
                                               double&      calculatedTangentialForceZ,
                                               double&      calculatedUnsymTangentialForceX,
                                               double&      calculatedUnsymTangentialForceY,
                                               double&      calculatedUnsymTangentialForceZ,
                                               double&      calculatedElem1AdditionalTorqueX,
                                               double&      calculatedElem1AdditionalTorqueY,
                                               double&      calculatedElem1AdditionalTorqueZ,
                                               double&      calculatedElem1UnsymAdditionalTorqueX,
                                               double&      calculatedElem1UnsymAdditionalTorqueY,
                                               double&      calculatedElem1UnsymAdditionalTorqueZ,
                                               double&      calculatedElem2AdditionalTorqueX,
                                               double&      calculatedElem2AdditionalTorqueY,
                                               double&      calculatedElem2AdditionalTorqueZ,
                                               double&      calculatedElem2UnsymAdditionalTorqueX,
                                               double&      calculatedElem2UnsymAdditionalTorqueY,
                                               double&      calculatedElem2UnsymAdditionalTorqueZ,
                                               double&      calculatedChargeMovedToElem1);

    private:
        std::string         m_brokenBondsFileName;
        std::ofstream       m_brokenBondsFile;
        bool                m_bondMakingHappening;
        double              m_requestedBondTime;
        double              m_bondingTime;
        double              m_bondingTimestep;
        CBondList           m_bonds;
        CBondParametersList m_bondParameters;

};

#endif
