#if !defined(cchargedparticle_h)
#define cchargedparticle_h

#include <string>

#include "Helpers.h"

#include "IPluginParticleBodyForceV2_1_0.h"

/**
 * This class provides an implementation of IPluginParticleBodyForceV2_1_0
 * that applies a force to particles.
 *
 * This plugin reads a config file from the same directory as
 * it is stored in.  The file format is as follows (1 line per value):
 *     plates_x
 *     plates_y
 *     plates_z
 *     frequency
 *     point
 *     particles
 *     point_position_x - X cordiante of point charge
 *     point_position_y - Y cordiante of point charge
 *     point_position_z - Z cordiante of point charge
 *     point_lower_x - X cordinate of start of plates AABB
 *     point_lower_y - Y cordinate of start of plates AABB
 *     point_lower_z - Z cordinate of start of plates AABB
 *     point_upper_x - X cordinate of end of plates AABB
 *     point_upper_y - Y cordinate of end of plates AABB
 *     point_upper_z - Z cordinate of end of plates AABB
 *
 * point_position is the position of a charged point source
 *
 * point_lower and point_upper define an Axis Aligned Bounding
 * box.  The top and bottom sides of which are used as charged
 * plates.
 */
class CChargedParticle : public NApiPbf::IPluginParticleBodyForceV2_1_0
{
    public:
        /**
         * Name of the preferences file to load bond information
         * from
         */
        static const std::string PREFS_FILE;

        /**
         * Value for Epsilon 0
         */
        static const double EPSILON0;

        /**
         * Constructor, does nothing
         */
        CChargedParticle();

        /**
         * Destructor, does nothing
         */
        virtual ~CChargedParticle();

        /**
         * Sets prefFileName to PREFS_FILE
         *
         *
         * Implementation of method from IPluginParticleBodyForceV2_1_0
         */
        virtual void getPreferenceFileName(char prefFileName[NApi::FILE_PATH_MAX_LENGTH]);

        /**
         * Returns true to indicate plugin is thread safe
         *
         * Implementation of method from IPluginParticleBodyForceV2_1_0
         */
        virtual bool isThreadSafe();

        /**
         * Reads config file
         *
         * Implementation of method from IPluginParticleBodyForceV2_1_0
         */
        virtual bool setup(NApiCore::IApiManager_1_0& apiManager,
                           const char                 prefFile[]);

        /**
         * Applies an ES force to particles
         *
         * Implementation of method from IPluginParticleBodyForceV2_1_0
         */
        virtual NApi::ECalculateResult externalForce(
                                               double       time,
                                               double       timestep,
                                               int          id,
                                               const char   type[],
                                               double       mass,
                                               double       volume,
                                               double       posX,
                                               double       posY,
                                               double       posZ,
                                               double       velX,
                                               double       velY,
                                               double       velZ,
                                               double       angVelX,
                                               double       angVelY,
                                               double       angVelZ,
                                               double       charge,
                                               const double orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* particlePropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* simulationPropData,
                                               double&      calculatedForceX,
                                               double&      calculatedForceY,
                                               double&      calculatedForceZ,
                                               double&      calculatedTorqueX,
                                               double&      calculatedTorqueY,
                                               double&      calculatedTorqueZ);

    private:
        double          m_point;
        double          m_particles;
        CSimple3DVector m_plates;
        CSimple3DPoint  m_pointPos;
        CSimple3DPoint  m_platesLower;
        CSimple3DPoint  m_platesUpper;
        double          m_freq;

};

#endif
