#include "CFluidDrag.h"

#include <fstream>

using namespace std;
using namespace NApi;
using namespace NApiCore;
using namespace NApiPbf;

const string CFluidDrag::PREFS_FILE = "fluid_drag_prefs.txt";

CFluidDrag::CFluidDrag()
{
    ;
}

CFluidDrag::~CFluidDrag()
{
    ;
}

void CFluidDrag::getPreferenceFileName(char prefFileName[FILE_PATH_MAX_LENGTH])
{
    // Copy in pref file name
    strcpy(prefFileName, PREFS_FILE.c_str());
}

bool CFluidDrag::isThreadSafe()
{
    // thread safe
    return true;
}

bool CFluidDrag::setup(NApiCore::IApiManager_1_0& apiManager,
                       const char                 prefFile[])
{
    ifstream prefsFile(prefFile);

    if(!prefsFile)
    {
        return false;
    }
    else
    {
        double pointX, pointY, pointZ;
        prefsFile >> pointX >> pointY >> pointZ;
        m_inlet = CSimple3DPoint(pointX,pointY,pointZ);

        prefsFile >> pointX >> pointY >> pointZ;
        m_outlet = CSimple3DPoint(pointX,pointY,pointZ);

        prefsFile >> m_diameter
                  >> m_fluidVeloctiy
                  >> m_dragCoefficient
                  >> m_density;

    }

    return true;
}

ECalculateResult CFluidDrag::externalForce(double       time,
                                           double       timestep,
                                           int          id,
                                           const char   type[],
                                           double       mass,
                                           double       volume,
                                           double       posX,
                                           double       posY,
                                           double       posZ,
                                           double       velX,
                                           double       velY,
                                           double       velZ,
                                           double       angVelX,
                                           double       angVelY,
                                           double       angVelZ,
                                           double       charge,
                                           const double orientation[9],
                                           NApiCore::ICustomPropertyDataApi_1_0* particlePropData,
                                           NApiCore::ICustomPropertyDataApi_1_0* simulationPropData,
                                           double&      calculatedForceX,
                                           double&      calculatedForceY,
                                           double&      calculatedForceZ,
                                           double&      calculatedTorqueX,
                                           double&      calculatedTorqueY,
                                           double&      calculatedTorqueZ)
{
    // This simulation consists of particles passing horizontally over
    // a fan which produces a vertical force on the particles. The
    // simulation assumes turbulent flow, and that the force is only applied
    // when the particle is in the direct path of the fan, within its diameter

    // Position of particle
    CSimple3DPoint nPos(posX, posY, posZ);

    // This part calculates the distance of the particle
    // from the central axis of the inlet
    CSimple3DVector stream       = m_outlet - m_inlet;
    CSimple3DVector inletToPoint = nPos - m_inlet;
    stream.normalise();

    // This gives the point on the inlet axis closest to the particle
    CSimple3DPoint onAxis = m_inlet + stream * inletToPoint.dot(stream);

    // Distance from axis:
    double nDistance = (nPos-onAxis).length();

    // Check if particle is within the stream
    // (if dist from axis is smaller than stream radius)
    if (nDistance < (m_diameter / 2.0))
    {
        // Equivalnet radius of particle (assuming spherical)
        double nRadius = pow( 3.0 * volume / ( 4.0 * PI ),1.0 / 3.0 );

        //cross sectional area of particle
        double nCSA =  PI * nRadius * nRadius;

        CSimple3DVector fluidVelocity = stream * m_fluidVeloctiy;
        CSimple3DVector particleVelocity(velX, velY, velZ);
        CSimple3DVector relativeVelocity = fluidVelocity - particleVelocity;

        //standard equation for drag past a bluff body
        CSimple3DVector F_total = relativeVelocity
                                  * 0.5
                                  * m_dragCoefficient
                                  * nCSA
                                  * m_density
                                  * relativeVelocity.length();

        calculatedForceX = F_total.dx();
        calculatedForceY = F_total.dy();
        calculatedForceZ = F_total.dz();
    }

    return eSuccess;
}