#include "CFieldQuery.h"
#include "IFieldApi_1_0.h"
#include "IFieldManagerApi_1_0.h"
#include "IApiManager_1_0.h"
#include <fstream>

using namespace std;
using namespace NApi;
using namespace NApiCore;
using namespace NApiPbf;

const string CFieldQuery::PREFS_FILE          = "field_query_prefs.txt";
const string CFieldQuery::FLUID_DRAG_PROPERTY = "Fluid drag";

CFieldQuery::CFieldQuery():
m_fieldManager(0), m_field(0)
{
    ;
}

CFieldQuery::~CFieldQuery()
{
    ;
}

void CFieldQuery::getPreferenceFileName(char prefFileName[FILE_PATH_MAX_LENGTH])
{
    // Copy in pref file name
    strcpy(prefFileName, PREFS_FILE.c_str());
}

bool CFieldQuery::isThreadSafe()
{
    // thread safe
    return true;
}

bool CFieldQuery::usesCustomProperties()
{
    // Use custom properties
    return true;
}

bool CFieldQuery::setup(NApiCore::IApiManager_1_0& apiManager,
                           const char                 prefFile[])
{
	//This is executed during the loading of the plugin
	//Open the preference file
    ifstream prefsFile(prefFile);

    if(!prefsFile)
    {
        return false;
    }
    else
    {
		//Reading in the parameters in the file
        prefsFile >> m_density
				  >> m_viscosity;
    }

    return true;
}

bool CFieldQuery::starting(NApiCore::IApiManager_1_0& apiManager)
{
	//This is executed at the start of the simulation run
	//Acquire API manager for field import
	m_fieldManager = dynamic_cast<IFieldManagerApi_1_0*>(apiManager.getApi(eFieldManager, 1, 0));

    if (0 == m_fieldManager)
    {
        return false;
    }

	//Acquire the required field data with respect to its specific name
    m_field = dynamic_cast<IFieldApi_1_0*>(m_fieldManager->getApi(eField, 1, 0, "CFD_field"));

    if (0 == m_field)
    {
        return false;
    }

    return true;
}

void CFieldQuery::stopping(NApiCore::IApiManager_1_0& apiManager)
{
	//This is executed at the end of the simulation run
    if (0 != m_fieldManager)
    {
        apiManager.release(m_fieldManager);
        m_fieldManager = 0;
    }

    if (0 != m_field)
    {
        apiManager.release(m_field);
        m_field = 0;
    }
}

ECalculateResult CFieldQuery::externalForce(double       time,
                                               double       timestep,
                                               int          id,
                                               const char   type[],
                                               double       mass,
                                               double       volume,
                                               double       posX,
                                               double       posY,
                                               double       posZ,
                                               double       velX,
                                               double       velY,
                                               double       velZ,
                                               double       angVelX,
                                               double       angVelY,
                                               double       angVelZ,
                                               double       charge,
                                               const double orientation[9],
                                               NApiCore::ICustomPropertyDataApi_1_0* particlePropData,
                                               NApiCore::ICustomPropertyDataApi_1_0* simulationPropData,
                                               double&      calculatedForceX,
                                               double&      calculatedForceY,
                                               double&      calculatedForceZ,
                                               double&      calculatedTorqueX,
                                               double&      calculatedTorqueY,
                                               double&      calculatedTorqueZ)
{
    //This simple example demonstrates the usage of field import in the API

    //Position of particle
    CSimple3DPoint nPos(posX, posY, posZ);

	double cfdFieldX, cfdFieldY, cfdFieldZ;

	//Get the vector for the particle centre
	if(m_field->queryVectorField(nPos.x(),nPos.y(),nPos.z(),8,eIdw,cfdFieldX,cfdFieldY,cfdFieldZ)==false)
	{
		cfdFieldX = 0;
		cfdFieldY = 0;
		cfdFieldZ = 0;
	}

	//Vector for the vector query
	CSimple3DVector cfdField(cfdFieldX,cfdFieldY,cfdFieldZ);

	//Relative velocity between the particle and the fluid
	CSimple3DVector particleVelocity(velX, velY, velZ);
	CSimple3DVector relativeVelocity = cfdField - particleVelocity;

	//Equivalent radius of particle (assuming spherical)
	double nRadius = pow( 3.0 * volume / ( 4.0 * PI ),1.0 / 3.0 );

	//cross sectional area of particle
	double nCSA =  PI * nRadius * nRadius;

	//Local Reynolds number
	double re = m_density * relativeVelocity.length() * 2.0 * nRadius / m_viscosity;

	if( !isZero(re) )
	{
		if( re <= 0.5)
		{
			m_dragCoefficient = 24.0 / re;
		}
		else if( re > 0.5 && re <= 1000.0)
		{
			m_dragCoefficient = 24.0 * ( 1.0 + 0.15 * pow(re,0.687) ) / re;
		}
		else
		{
			m_dragCoefficient = 0.44;
		}

		//standard equation for drag past a bluff body
		CSimple3DVector F_total = relativeVelocity
									  * 0.5
									  * m_dragCoefficient
									  * nCSA
									  * m_density
									  * relativeVelocity.length();

		calculatedForceX = F_total.dx();
		calculatedForceY = F_total.dy();
		calculatedForceZ = F_total.dz();

		// Custom properties to record the drag force

		// First reset the custom properties itself
        //
        // As we can only apply delta values to the properties
        // we apply the negative of their current values to zero them
        //
        // This would cause real problems if a second plugin
        // tried to do the same thing
        
		double* m_fluidDragDelta = particlePropData->getDelta(FLUID_DRAG_PROPERTY.c_str());
		const double* m_fluidDragValue = particlePropData->getValue(FLUID_DRAG_PROPERTY.c_str());		
		
		m_fluidDragDelta[0] -= m_fluidDragValue[0];
		m_fluidDragDelta[1] -= m_fluidDragValue[1];
		m_fluidDragDelta[2] -= m_fluidDragValue[2];

        // Then add in the new values
		m_fluidDragDelta[0] += calculatedForceX;
		m_fluidDragDelta[1] += calculatedForceY;
		m_fluidDragDelta[2] += calculatedForceZ;

		//Gravity
		CSimple3DVector gravity(0.0,-9.81,0.0);

		//Buoyancy force
		CSimple3DVector F_buoyancy = gravity * volume * m_density;

		calculatedForceX -= F_buoyancy.dx();
		calculatedForceY -= F_buoyancy.dy();
		calculatedForceZ -= F_buoyancy.dz();
	}

    return eSuccess;
}

unsigned int CFieldQuery::getNumberOfRequiredProperties(
                                const NApi::EPluginPropertyCategory category)
{
	if (eParticle == category)
	{
		return 1;
	}
	else
    {
        return 0;
    }
}

bool CFieldQuery::getDetailsForProperty(
                             unsigned int                    propertyIndex,
                             NApi::EPluginPropertyCategory   category,
                             char                            name[NApi::CUSTOM_PROP_MAX_NAME_LENGTH],
                             NApi::EPluginPropertyDataTypes& dataType,
                             unsigned int&                   numberOfElements,
                             NApi::EPluginPropertyUnitTypes& unitType)
{

    if (0 == propertyIndex)
    {
        strcpy(name, FLUID_DRAG_PROPERTY.c_str());
        dataType         = eDouble;
        numberOfElements = 3;
        unitType         = eForce;
        return true;
    }
    else
    {
        return false;
    }
}